#[derive(Debug, PartialEq, Eq, Clone, Copy)]
pub enum Error {
    /// Failed to process the sent instruction packet
    Result,
    /// Undefined instruction has been used
    Instruction,
    /// CRC of the sent packet does not match
    Crc,
    /// Data to be written in the corresponding address is outside the range of the minimum /
    /// maximum value
    DataRange,
    /// Attempt to write data that is shorter than the data length of the corresponding address
    DataLength,
    /// Data to be written in the corresponding address is outside of the limit value
    DataLimit,
    /// Access error (write to read only, read from write only, write to rom in rom lock)
    Access,
    /// There is some hardware issue with the device. Check hardware error status for details.
    Hardware,
    /// An unknown error code was returned
    Unknown,
    /// The buffer is to small
    BufferSize,
}

impl Error {
    /// Converts the error byte from a status packet into a rust result.
    const fn from_byte(byte: u8) -> Result<(), Self> {
        match byte & 0x7F {
            0 if byte & 0x80 == 0 => Ok(()),
            0 => Err(Self::Hardware),
            1 => Err(Self::Result),
            2 => Err(Self::Instruction),
            3 => Err(Self::Crc),
            4 => Err(Self::DataRange),
            5 => Err(Self::DataLength),
            6 => Err(Self::DataLimit),
            7 => Err(Self::Access),
            _ => Err(Self::Unknown), // This shouldn't be possible
        }
    }
}
